function w4DvarML()
%% Weak 4d Variational for Morris-Lecar
% Produces example figure for initializing with HOPF regime parameters for
% estimating parameters associated with SNIC regime.
%%% Preliminary setup
obsNoise=.01; sd=4;rng(sd);
T=2000; dT=.1; dt=.1; nn = fix(dT/dt);
truep = [.067,4,12,17.4,120,-84,-60,8,2,-1.2,18,20,100];
guessp= [.04,4,2,30,120,-84,-60,8,2,-1.2,18,20,100];
SimName='tSNICgHOPF';
EstSet={'V','n','phi','gCa','V3','V4','gK','gL','V1','V2'};
dx = length(EstSet); Ntheta=8;
%%% Define weightings for cost function (inv(Q), inv(R))
measStdDev=1;% observational noise variance is StDev^2
Rinv=inv(measStdDev^2);
% The dynamic range of v is larger than n by orders of magnitude. Need to
% weight deviations appropriately. Also need to choose ratio of measurement
% to model noise.
weightScales=10.^[-4:5]; weightIndex = 7;
weightScale=weightScales(weightIndex);
Qinv=weightScale*Rinv*diag([1 10000]);
% Presumes each time series component of the system has equal weighting
% For simplicity of implementation, Q,R taken as diagonal matrix
%%% Generate Data
x00=[-4 .14]';
Ttotal=5*T;
parameters=[ones(Ttotal+1,1)*truep];
x0=generateMLME(x00,Ttotal,parameters,dT);
x0=generateMLME(x0(:,end),Ttotal,parameters,dT);
parametersGuess=[ones(Ttotal+1,1)*guessp];
%%% guesses
% Set initial guesses.
Vtruth=x0(1,:); ntruth=x0(2,:);
signalStdev=std(Vtruth);
y=Vtruth+ obsNoise*(signalStdev.*randn(size(Vtruth)));
% Guess n to be integration of guessed model forced with observations
x02=generateMLMEVforced(x00(2),Ttotal,parametersGuess,dT,y);
nguess=x02(1,:);
% Used more data than needed, pick subset to estimate over
windowInds=2*T+1:3*T+1;
Vobswindow=y(windowInds);
nguesswindow=nguess(windowInds);
InitGuess=zeros(2,T+1);
InitGuess(1,:)=Vobswindow;
InitGuess(2,:)=nguesswindow;
InitGuess=[InitGuess(1,:)'; InitGuess(2,:)'];
pEstInds=[1 2 3 4 8 9 10 11];
InitGuess =[InitGuess; guessp(pEstInds)'];
%%%
% InitGuess is a stack of guesses for system components
% InitGuess = [V_1 V_2... V_{J+1} n_1 n_2 ... n_{J+1}, phi, gCa, V3, V4, gK , gL , V1, V2]';
% Next set the bounds.
LB = zeros(size(InitGuess)); RB=LB;
LB(1:T+1)=-100; LB(T+2:end-Ntheta)=0;
RB(1:T+1)=100;  RB(T+2:end-Ntheta)=1;
LB(end-Ntheta+1:end)= [0 0 -20 .1 0 0 -10 0.1]';
RB(end-Ntheta+1:end)= [1 10 20 35 10 5 20 35]';
%%% Construct an output file/function to save data-stream
ClockTime=fix(clock);
ClockString=[num2str(ClockTime(1)),'_',num2str(ClockTime(2)),'_',num2str(ClockTime(3))];
FileName=['w4dvar',SimName,'obs',num2str(obsNoise),...
    'wscale',num2str(weightIndex),'T',ClockString];
FileName=strrep(FileName,'.','pt'); % same output during iteration cycle
myOutputFun =@(x,optimValues,state)w4dvarOutputFun(x,optimValues,state,[FileName,'.txt']);
%%% Options can be changed based on performance, preferences, etc.
options1 = optimoptions('fmincon','Display','iter',...
    'MaxFunctionEvaluations',4e8,'MaxIterations',1e5,...
    'HessianApproximation','lbfgs','InitBarrierParam',1e-3,...% limited memory hessian approx
    'OutputFcn',@(x,optimValues,state)myOutputFun(x,optimValues,state));
%%% Run optimization
[xest,fval,exitflag,output] = fmincon(@(u)w4dvarobjfun(u,Vobswindow,Qinv,...
    Rinv,@MLstatefctw4dvarModEuler,T,truep,Ntheta,dt,nn,dx),InitGuess,...
    [],[],[],[],LB,RB,[],options1);
%save(FileName) %(uncomment to save all data at end in .mat)
end
function [fout]=w4dvarobjfun(xbold,y,Qinv,Rinv,F,T,truep,Ntheta,dt,nn,dx)
% The following code accepts as arguments:
%
% xbold : vector of N*(T+1) + P dimensions, for N being total number of
% states in dynamical model (2 for Morris Lecar)
%
% y : observation. Currently set to operate on an observation operator
% which is a simple projection onto the first system component i.e.
% H_{1,1} = 1, H_{j,k} = 0 otherwise.
%
% Qinv : inverse covariance matrix associated with model error.
% inv(smaller values of covariance) imply larger weights towards those
% terms in cost functional. Taken to be a diagonal matrix.
%
% Rinv : inverse covariance matrix associated with measurement
% noise
%
% F : discrete map representation of model. for our purposes, F is a
% numerical integration scheme applied to a model in question e.g.
% Morris-Lecar. Currently, using modified Euler set up for autonomous system.
%
% T : T+1 points in time
%
% truep : passing parameters to model. for those which will be
% determined from the scheme, their truep counterparts will not be
% used, but their component in xbold will. These are at the end of xbold.
theta=xbold(end-Ntheta+1:end);
v=xbold(1:T+1);
n=xbold(T+2:end-Ntheta);
u = [reshape(v,1,T+1); reshape(n,1,T+1)];
MeasurementMismatch=0; ModelMismatch=0;
%%% Compute F(x_k; \theta) for all k
fwdmapj=F(u,theta,truep,dt,nn,dx);
%%% Model term
for l=1:2
    ModelMismatch=ModelMismatch+...
        1/2*sum(Qinv(l,l)*((u(l,2:end)-fwdmapj(l,1:end-1))).^2); 
end
%%% Measurement term
MeasurementMismatch=MeasurementMismatch+1/2*sum(Rinv*((y(1,:)-u(1,:))).^2);
fout=MeasurementMismatch+ModelMismatch;
end
function r=MLstatefctw4dvarModEuler(x,theta,p,dt,nn,dx)
Ntheta=length(theta);
xnl=zeros(dx,size(x,2));
xnl(1:2,:)=x;
xnl(3:end,:)=repmat(reshape(theta,Ntheta,1),1,size(x,2));
for n=1:nn
    k1=dt*fXaug(0,xnl,p);
    k2=dt*fXaug(0,xnl+k1,p);
    r=xnl+k1./2 + k2./2;
end
end
function x=generateMLMEVforced(x0,T,p,dt,V)
x(:,1)=x0;
for j=1:T
k1=dt*fXVforced(0,x(:,j),p(j,:),V(j));
k2=dt*fXVforced(0,x(:,j)+k1,p(j+1,:),V(j+1));
x(:,j+1)=x(:,j)+k1./2+k2./2;
end
end
function dy = fXVforced(t,y,p,V)
phi=p(1);
V3=p(3);
V4=p(4);
n = y(1,:);
dy=zeros(size(y));
dy(1,:)=fn(V,n,phi,V3,V4);
end
